#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "camclaycoup.h"
#include "global.h"
#include "vecttens.h"
#include "intpoints.h"
#include "matrix.h"
#include "vector.h"
#include "tensor.h"
#include "elastisomat.h"



/**
  This constructor inializes attributes to zero values.
*/
camclaycoup::camclaycoup (void)
{
  m = 0.0;
  lambda = 0.0;
  kappa = 0.0;
  c1_tilde = 0.0;
  c2 = 0.0;
  ks = 0.0;
}



/**
  This destructor is only for the formal purposes.
*/
camclaycoup::~camclaycoup (void)
{

}



/**
  This function reads material parameters from the opened text file given
  by the parameter in.

  @param in - pointer to the opned text file

*/
void camclaycoup::read (XFILE *in)
{
  xfscanf (in, "%k%lf%k%lf%k%lf", "m", &m, "lambda", &lambda, "kappa", &kappa);
  xfscanf (in, "%k%lf%k%lf%k%lf", "c1_tilde", &c1_tilde, "c2", &c2, "ks", &ks);
  sra.read (in);
}

/**
   This function initializes material model data
   with respect of consistency parameter gamma.

   @param ipp - integration point number
   @param ido - index of internal variables for given material in the ipp other array

   12/06/2012 TKo
*/
void camclaycoup::initval(long ipp, long ido)
{
  double v_ini, i1s, v_pc0, v_lambda1;
  double v_kappa1, p1, pc_0;
  long i, ncompstr = Mm->ip[ipp].ncompstr;
  vector epse(ncompstr), sig(ncompstr);
  matrix d(ncompstr, ncompstr), sigt(3,3);
  
  if (Mm->ip[ipp].eqother[ido+ncompstr+1] == 0.0)  // actual value of p_c was not set
  {
    // initial value of specific volume under small pressure p_1
    v_kappa1 = Mm->ic[ipp][0];
    // initial reference presure
    p1       = Mm->ic[ipp][1];
    // initial effective preconsolidation pressure
    pc_0     = Mm->ic[ipp][2];

    //  initial value of elastic strain
    for (i=0; i<ncompstr; i++)
      epse[i] = Mm->ic[ipp][3+i];

    //  elastic stiffness matrix
    Mm->elmatstiff (d,ipp);
    // initial value of stress
    mxv(d, epse, sig);
    vector_tensor (sig,sigt,Mm->ip[ipp].ssst,stress);
    // initial value of mean stress
    i1s = first_invar(sigt)/3.0;

    // specific volume for initial effective preconsolidation pressure
    v_pc0 = v_kappa1 - kappa*log(pc_0/p1);
    // actual value of p_c = initial effective preconsolidtaion pressure
    Mm->ip[ipp].eqother[ido+ncompstr+1] = pc_0;
    // original specific volume before any loading
    Mm->ip[ipp].eqother[ido+ncompstr+3] = v_lambda1 = v_pc0 + lambda*log(pc_0/p1);
    // specific volume for intial stress state
    Mm->ip[ipp].eqother[ido+ncompstr+4] = v_ini     = v_pc0 + kappa*log(pc_0/i1s);
  }

  return;
}


/**
   This function computes the value of yield functions.

   @param sig - stress tensor
   @param q   - %vector of hardening parameter

   @retval The function returns value of yield function for the given stress tensor

   25.3.2002
*/
double camclaycoup::yieldfunction (matrix &sig, vector &q)
{
  double i1s,j2s,f;

  matrix dev(3,3);
  deviator (sig,dev);

  // q[0] = \bar{p}_c
  // q[1] = p_s = k * s
  // f(\sigma, s, p_c) = q^2/(M^2) + (p-p_s)*(p-\bar{p}_c)
  i1s = first_invar (sig)/3.0;
  j2s = second_invar (dev);
  f = j2s/(m*m) + (i1s - q[1]) * (i1s - q[0]);
  
  return f;
}



/**
   This function computes derivatives of as-th yield function
   with respect of vector sigma.

   @param sig - stress tensor
   @param q   - %vector of the hardening parameter
   @param dfds - %matrix where the resulting derivatives are stored


   4.1.2002
*/
void camclaycoup::deryieldfsigma (matrix &sig, vector &q, matrix &dfds)
{
  double volume,i1s;
  matrix dev(3,3);

  i1s = first_invar (sig)/3.0;

  deviator (sig,dfds);

  // q[0] = \bar{p}_c
  // q[1] = p_s = k * s
  cmulm(1.0/(m*m), dfds);
  volume=1.0/3.0*(2.0*i1s - q[0] - q[1]);
  dfds[0][0]+=volume;
  dfds[1][1]+=volume;
  dfds[2][2]+=volume;  
}



/**
   The function computes the second derivatives of yield function
   with respect of stress tensor sigma.

   @param ddfds - tensor of the 4-th order where the are derivatives stored

   19.12.2002
*/
void camclaycoup::dderyieldfsigma (matrix &ddfds)
{
  fillm(0.0, ddfds);
  ddfds[0][0] = ddfds[1][1] = ddfds[2][2] = 2.0/(3.0*m*m) + 2.0/9.0;
  ddfds[0][1] = ddfds[0][2] = ddfds[1][0] = ddfds[1][2] = -1.0/(3.0*m*m) + 2.0/9.0;
  ddfds[2][0] = ddfds[2][1] = ddfds[0][1];
  ddfds[3][3] = 1.0/(m*m);
  ddfds[4][4] = 1.0/(m*m);
  ddfds[5][5] = 1.0/(m*m);
}



/**
   The function computes derivatives of plastic potential function
   with respect of vector sigma.

   @param sig - stress tensor
   @param q   - %vector of the hardening parameters
   @param dgds - %matrix where the resulting derivatives are stored
*/
void camclaycoup::derpotsigma (matrix &sig, vector &q, matrix &dgds)
{
  deryieldfsigma (sig, q, dgds);
}



/**
   This function computes derivatives of as-th yield function
   with respect of vector of hradening parameters.

   @param sig - stress tensor
   @param q   - %vector of the hardening parameters
   @param dfds - %vector where the resulting derivatives are stored


   4.1.2002
*/
void camclaycoup::deryieldfq(matrix &sig, vector &q, vector &dfq)
{
  //     df
  // ---------- = dfq[0]
  // d\bar{p}_c
  dfq[0] = -first_invar (sig)/3.0 + q[1];
  // df
  // --- = dfq[1]
  // p_s
  dfq[1] = -first_invar (sig)/3.0 + q[0];

  return;
}



/**
  The function computes the second derivatives of yield function
  with respect to hradening parameters.

  @param dfds - %matrix, where the resulting derivatives are stored


  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::deryieldfdqdq(matrix &dfq)
{  
  dfq[0][0] = 0.0;
  dfq[0][1] = 1.0;
  dfq[1][0] = 1.0;
  dfq[1][1] = 0.0;

  return;
}



/**
  The function computes the second derivatives of yield function
  with respect to stresses.

  @param dfds - tensor, where the resulting derivatives are stored.
                size of dfds = (6,number_of_hardening_param)


  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::deryieldfdsdq(matrix &dfdsdqt)
{  
  dfdsdqt[0][0] = dfdsdqt[1][0] = dfdsdqt[2][0] = -1.0/3.0;
  dfdsdqt[0][1] = dfdsdqt[1][1] = dfdsdqt[2][1] = -1.0/3.0;
  return;
}



/**
   The function computes derivatives of hardening paramters
   with respect of consistency parameter gamma.

   @param ipp - integration point number
   @param ido - index of internal variables for given material in the ipp other array
   @param sig - stress tensor
   @param qtr  - %vector of hardening variables
   @param dqdg - %matrix where the resulting derivatives are stored


  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::der_q_gamma(long ipp, long ido, matrix &sig, vector &qtr, vector &dqdg)
{
  double v_ini, v_lambda1, i1s;
  double p_atm = 101325.0;
  double s, sr, fs, b, c, c1, ksi;
  long ncompstr = Mm->ip[ipp].ncompstr;

  // mean stress
  i1s = first_invar (sig)/3.0;
  // original specific volume before any loading
  v_lambda1 = Mm->ip[ipp].eqother[ido+ncompstr+3];
  // specific volume for intial stress state
  v_ini = Mm->ip[ipp].eqother[ido+ncompstr+4];

  //s  = Mm->nonmechq[ipp][4];  // suction pressure s
  //sr = Mm->nonmechq[ipp][5];  // degree of saturation S_r
  s  = Mm->givenonmechq(suction, ipp);  // suction pressure s
  sr = Mm->givenonmechq(saturation_degree, ipp);  // degree of saturation S_r

  // suction function:
  //
  //                  s / p_{atm}
  // f(s) = 1 + -----------------------
  //            10.7 + 2.4(s / p_{atm})
  //
  fs = 1.0 + (s/p_atm) / (10.7 + 2.4*(s/p_atm)); 

  // bonding variable:
  //
  // \ksi = f(s) (1 - S_r)
  //
  ksi = (1-sr)*fs;    

  //            \tilde{c}_1
  // c_1 = --------------------
  //           1
  //        -------------- + 1
  //        v_{\lambda1}-1
  c1 = c1_tilde/(1.0/(v_lambda1-1.0)+1.0);


  //
  // c(\ksi) = 1 - c_1 [1-\exp(c_2 \ksi)]
  //
  c = 1.0 - c1*(1.0-exp(c2*ksi));

  //
  //                \lambda - \kappa
  // b(\ksi) = --------------------------
  //            \lambda c(\ksi) - \kappa
  //
  b = (lambda - kappa)/(lambda*c - kappa);


  // q[0] = \bar{p}_c = -\exp[a(\ksi)] (-p_c)^{b(\ksi)}
  //               
  // p_c  = p_1 exp[(v_{\kappa} - v_{\lambda_1})/(\kappa - \lambda)]
  //
  // v_{\kappa}    = v_{ini} (1+\eps_v) + \kappa  \ln(p/p_1) => depends on p => depends on gamma
  // 
  // v_{ini}       = v_{p_{c0}}         + \kappa  \ln(p_{c0}/p_{ini}) = const
  // v_{\lambda_1} = v_{p_{c0}}         + \lambda \ln(p_{c0}/p_1)     = const
  // v_{p_{c0}}    = v_{\kappa_1}       - \kappa  \ln(p_{c0}/p_1)     = const
  // 
  // d \bar{p}_c                                              d p_c    
  // ----------- = \exp[a(\ksi)] b(\ksi)(-p_c)^{b(\ksi)-1} ----------, 
  //  d \gamma                                               d \gamma  
  //                             
  //   d p_c               v_ini
  // ---------- = p_c ----------------- (2p - \bar{p}_c - p_s),
  //  d \gamma        \kappa - \lambda
  //
  //  d \bar{p}_c              b(\ksi) v_ini                              
  //  ----------- = -\bar{p}_c ---------------- (2p - \bar{p}_c - p_s)
  //   d \gamma               \kappa - \lambda                       
  //
  //
  dqdg[0] = -qtr[0] * (b*v_ini)/(kappa-lambda) * (2.0*i1s-qtr[0]-qtr[1]);


  // q[1] = p_s = k*s
  //
  //    d p_s
  // ----------- = 0.0 because p_s is constant for the given strain level
  //  d \gamma
  //
  dqdg[1] = 0.0;

  return;
}



/**
  This function computes plastic modulus.

  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array
  @param sig - stress tensor
  @param qtr - %vector of hardening parameters

  @retval The function returns value of the plastic modulus.

  Created by Tomas Koudelka,  09.2012
*/
double camclaycoup::plasmodscalar(long ipp, long ido, matrix &sig, vector &qtr)
{
  double ret;
  vector dfq(2);
  vector dqg(2);

  deryieldfq(sig, qtr, dfq);
  der_q_gamma(ipp, ido, sig, qtr, dqg);
  scprd(dfq, dqg, ret);

  return -ret;
}



/**
  This function computes new value of the hardening parameter q.

  @param ipp  - integration point pointer
  @param ido  - index of internal variables for given material in the ipp other array
  @param eps  - %vector of the attained strains
  @param epsp - %vector of the attained plastic strains
  @param sig  - attained stress tensor
  @param ssst - stress/strain state parameter (for the used vector_tensor function)
  @param q    - %vector of the hardening parameters

  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::updateq(long ipp, long ido, vector &eps, vector &epsp, matrix &sig, vector &q)
{
  matrix epst(3,3);
  vector depsp(epsp.n);
  double v_kappa1, v_lambda1, p1, depsvp, v_ini, pc_new;
  double v_pc, p_atm = 101325.0;
  double s, sr, fs, a, b, c, c1, ksi;
  strastrestate ssst = Mm->ip[ipp].ssst;
  long ncompstr = Mm->ip[ipp].ncompstr;
  long i;

  // initial value of specific volume under small pressure p_1
  v_kappa1 = Mm->ic[ipp][0];
  // initial reference presure
  p1 = Mm->ic[ipp][1];
  // original specific volume before any loading
  v_lambda1 = Mm->ip[ipp].eqother[ido+ncompstr+3];
  // specific volume for intial stress state
  v_ini     = Mm->ip[ipp].eqother[ido+ncompstr+4];

//  Original version of pc calculation - problems with tension
//  vk = v_ini*(1+epsv)+kappa*log(i1s/p1);
//  pc_new = p1 * exp((vk-v_lambda1)/(kappa-lambda));
//
//  Alternative version of pc calculation - 'less' accurate in benchmarks
//  pc_new = p1 * exp((v_kappa1 - v_lambda1 + v_ini*epsv)/(kappa-lambda));
//  vk = v_ini*(1+epsv)+kappa*log(i1s/p1);

  for (i=0; i<ncompstr; i++)
    depsp[i] = epsp[i] + Mm->ic[ipp][3+i];
  vector_tensor (depsp, epst, ssst, strain);
  depsvp = first_invar(epst);
  pc_new = p1 * exp((v_kappa1 - v_lambda1 + v_ini*depsvp)/(kappa-lambda));

  //  For the partially saturated soil, 
  //  the effective preconsolidation pressure \bar{p}_c is related to the saturated 
  //  preconsolidation pressure p_c according to bonding variable \ksi
  //
  // \bar{p}_c = -exp[a(\ksi)] (-p_c)^{b(\ksi)}
  //
  //
 
  // specific volume for the actual preconsolidation pressure:
  //
  // v_{pc} = v_{\lambda1} - \lambda \ln(p_c/p_1)
  v_pc = v_lambda1 - lambda * log(pc_new/p1);

  //s  = Mm->nonmechq[ipp][4];  // suction pressure s
  //sr = Mm->nonmechq[ipp][5];  // degree of saturation S_r
  s  = Mm->givenonmechq(suction, ipp);  // suction pressure s
  sr = Mm->givenonmechq(saturation_degree, ipp);  // degree of saturation S_r

  // suction function:
  //
  //                  s / p_{atm}
  // f(s) = 1 + -----------------------
  //            10.7 + 2.4(s / p_{atm})
  //
  fs = 1.0 + (s/p_atm) / (10.7 + 2.4*(s/p_atm)); 

  // bonding variable:
  //
  // \ksi = f(s) (1 - S_r)
  //
  ksi = (1-sr)*fs;    

  //            \tilde{c}_1
  // c_1 = --------------------
  //           1
  //        -------------- + 1
  //        v_{\lambda1}-1
  c1 = c1_tilde/(1.0/(v_lambda1-1.0)+1.0);


  //
  // c(\ksi) = 1 - c_1 [1-\exp(c_2 \ksi)]
  //
  c = 1.0 - c1*(1.0-exp(c2*ksi));

  //
  // a(\ksi) = v_ini v_{lambda_1} [c(\ksi)-1]/[\lambda c(\ksi) - \kappa)
  //
  a = v_ini*v_lambda1*(c - 1.0)/(lambda*c - kappa);

  //
  //                \lambda - \kappa
  // b(\ksi) = --------------------------
  //            \lambda c(\ksi) - \kappa
  //
  b = (lambda - kappa)/(lambda*c - kappa);

  //
  // effective preconsolidation pressure:
  //
  // q[0] = \bar{p}_c = -exp[a(\ksi)] (-p_c)^{b(\ksi)}
  //
  q[0] = -exp(a)*pow(-pc_new, b);
  

  // apparent adhesion due to suction
  // (right intersection of the yield surface with p axis):
  //
  // q[1] = p_s = k s
  //
  q[1] = ks * s;
  return;
}



/**
  This function computes material stiffnes matrix.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array

  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::matstiff (matrix &d, long ipp,long ido)
{
  double zero=1.0e-20;

  if (Mp->nlman->stmat==1)
  {
    //  initial elastic matrix
    Mm->elmatstiff (d,ipp);
  }
  if (Mp->nlman->stmat==2)
  {
    //  consitent tangent stiffness matrix
    matrix e(d.m, d.n), de(d.m, d.n), ddfdst(6,6), tmp(d.m, d.n), ddfds(d.m,d.n);
    long n = Mm->ip[ipp].ncompstr, i;    
    double gamma = Mm->ip[ipp].other[ido+n];
    for (i = 0; i < d.m; i++)
      e[i][i] = 1.0;
    Mm->elmatstiff (de,ipp);
    dderyieldfsigma (ddfdst);
    //  matrix representation of the fourth order tensor
    tensor4_matrix (ddfds,ddfdst,Mm->ip[ipp].ssst);
    cmulm(gamma, de, d);
    mxm(d, ddfds, tmp);
    addm(e, tmp, d);
    invm(d, tmp,zero);
    mxm(tmp, de, d);
  }
}



/**
  This function computes stresses at given integration point ipp,
  depending on the reached strains.
  The cutting plane algorithm is used. The stress and the other attribute of
  given integration point is actualized.

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::nlstresses (long ipp, long im, long ido)
  //
{
  long i,ni,ncomp=Mm->ip[ipp].ncompstr;
  double gamma,err;
  vector epsn(ncomp),epsp(ncomp),q(2);
  vector epsa(ncomp),sig(ncomp),dfds(ncomp),dgds(ncomp);
  matrix d(ncomp,ncomp),sigt(3,3),dfdst(3,3),dgdst(3,3);
  matrix epst(3,3), epspt(3,3);
  matrix dev(3,3);
  double i1s, j2s, epsv, epsvp;

  //  initial values
  for (i=0; i<ncomp; i++)
  {
    epsn[i] = Mm->ip[ipp].strain[i];
    epsp[i] = Mm->ip[ipp].eqother[ido+i];
    epsp[i] += Mm->ic[ipp][3+i];
  }
  gamma=Mm->ip[ipp].eqother[ido+ncomp];
  q[0] = Mm->ip[ipp].eqother[ido+ncomp+1];  // effective preconsolidation pressure \bar{p}_c
  q[1] = Mm->ip[ipp].eqother[ido+ncomp+2];  // apparent adhesion  p_s = k * s

  //  stress return algorithm
  switch(sra.give_tsra ())
  {
    case cp:
      ni=sra.give_ni ();
      err=sra.give_err ();
      Mm->cutting_plane (ipp, im, ido, gamma, epsn, epsp, q, ni, err);
      break;
    case gsra:
      ni=sra.give_ni ();
      err=sra.give_err ();
      Mm->newton_stress_return (ipp, im, ido, gamma, epsn, epsp, q, ni, err);
      break;
    default:
      print_err("wrong type of stress return algorithm is required", __FILE__, __LINE__, __func__);
      abort ();
  }
  //  new data storage
  for (i=0; i<ncomp; i++){
    epsp[i] -= Mm->ic[ipp][3+i];    
    Mm->ip[ipp].other[ido+i]=epsp[i];
  }
  Mm->ip[ipp].other[ido+ncomp]=gamma;
  Mm->ip[ipp].other[ido+ncomp+1]=q[0]; // the first hardening parameter \bar{p}_c
  Mm->ip[ipp].other[ido+ncomp+2]=q[1]; // the second hardening parameter p_s

  Mm->givestress(0, ipp, sig);
  vector_tensor (sig,sigt,Mm->ip[ipp].ssst,strain);
  vector_tensor (epsn,epst,Mm->ip[ipp].ssst,strain);
  vector_tensor (epsp,epspt,Mm->ip[ipp].ssst,strain);
  deviator (sigt,dev);
  i1s = first_invar (sigt)/3.0;
  j2s = sqrt(second_invar (dev));
  epsv = first_invar (epst);
  epsvp = first_invar (epspt);
  Mm->ip[ipp].other[ido+ncomp+5] = i1s;  
  Mm->ip[ipp].other[ido+ncomp+6] = j2s;  
  Mm->ip[ipp].other[ido+ncomp+7] = epsv;  
  Mm->ip[ipp].other[ido+ncomp+8] = epsvp;

}



/**
  This function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::updateval (long ipp,long ido)
{
  long i,n = Mm->ip[ipp].ncompstr;

  for (i=0;i<n;i++){
    Mm->ip[ipp].eqother[ido+i]=Mm->ip[ipp].other[ido+i];
  }
  Mm->ip[ipp].eqother[ido+n]   = Mm->ip[ipp].other[ido+n];       // gamma
  Mm->ip[ipp].eqother[ido+n+1] = Mm->ip[ipp].other[ido+n+1];     // p_c (hardening parameter - effective preconsolidation pressure) 
  Mm->ip[ipp].eqother[ido+n+2] = Mm->ip[ipp].other[ido+n+2];     // p_s (hardening parameter - apparent adhesion) 
  //
  // Following three values of the specific volume are constant and they 
  // are stored directly in the eqother array during the intial step
  //
  //  Mm->ip[ipp].eqother[ido+n+3] = Mm->ip[ipp].other[ido+n+3];     // v_lambda1
  //  Mm->ip[ipp].eqother[ido+n+4] = Mm->ip[ipp].other[ido+n+4];     // v_ini
  //
  Mm->ip[ipp].eqother[ido+n+5] = Mm->ip[ipp].other[ido+n+5];     // i1s;  
  Mm->ip[ipp].eqother[ido+n+6] = Mm->ip[ipp].other[ido+n+6];     // j2s;  
  Mm->ip[ipp].eqother[ido+n+7] = Mm->ip[ipp].other[ido+n+7];     // epsv;  
  Mm->ip[ipp].eqother[ido+n+8] = Mm->ip[ipp].other[ido+n+8];     // epsvp;  
  
  //data storing for coupled problems 12/06/2012 TKr
/*
  if(Mm->nonmechq != NULL){
    Mm->nonmechq[ipp][1] = Mm->ip[ipp].eqother[ido+n+5];                                            //i1s
    Mm->nonmechq[ipp][2] = Mm->ip[ipp].eqother[ido+n+3] - 1.0;                                      //e_lambda1
    Mm->nonmechq[ipp][3] = Mm->ip[ipp].eqother[ido+n+4]*(1.0 + Mm->ip[ipp].eqother[ido+n+7]) - 1.0; //e_ini;
  }
*/
}



/**
  Function returns irreversible plastic strains.

  @param ipp   - integration point number in the mechmat ip array.
  @param ido   - index of the first internal variable for given material in the ipp other array
  @param epsp  - %vector of irreversible strains
 
  Returns vector of irreversible strains via parameter epsp

  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::giveirrstrains (long ipp, long ido, vector &epsp)
{
  long i;
  for (i=0;i<epsp.n;i++)
    epsp[i] = Mm->ip[ipp].eqother[ido+i];
}



/**
  This function extracts consistency parametr gamma for the reached equilibrium state
  from the integration point other array.

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  @retval The function returns value of consistency parameter.

  Created by Tomas Koudelka,  09.2012
*/
double camclaycoup::give_consparam (long ipp,long ido)
{ 
  long ncompstr;
  double gamma;

  ncompstr=Mm->ip[ipp].ncompstr;
  gamma = Mm->ip[ipp].eqother[ido+ncompstr];

  return gamma;
}



/**
  The function extracts preconsolidation pressure p_c for the attained equilibrium state
  from the integration point eqother array.

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  @retval The function returns value of preconsolidation pressure.

  Created by Tomas Koudelka, 8.10.2013  
*/
double camclaycoup::give_preconspress(long ipp, long ido)
{
  long ncompstr;
  double pc;

  ncompstr=Mm->ip[ipp].ncompstr;
  pc = Mm->ip[ipp].eqother[ido+ncompstr+1];
  
  return pc;
}



/**
  The function extracts virgin porosity e_lambda1 for the attained equilibrium state
  from the integration point eqother array.

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  @retval The function returns value of virgin porosity e_lambda1.

  Created by Tomas Koudelka, 8.10.2013  
*/
double camclaycoup::give_virgporosity(long ipp, long ido)
{
  long ncompstr;
  double e_lambda1;

  ncompstr=Mm->ip[ipp].ncompstr;
  e_lambda1 = Mm->ip[ipp].eqother[ido+ncompstr+3]-1.0;
  
  return e_lambda1;
}



/**
  The function extracts initial porosity e_ini for the attained equilibrium state
  from the integration point eqother array.

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array

  @retval The function returns value of initial porosity e_ini.

  Created by Tomas Koudelka, 8.10.2013  
*/
double camclaycoup::give_iniporosity(long ipp, long ido)
{
  long ncompstr;
  double e_ini;

  ncompstr=Mm->ip[ipp].ncompstr;
  e_ini = Mm->ip[ipp].eqother[ido+ncompstr+4]-1.0;
  
  return e_ini;
}



/**
  The funtion marks required non-mechanical quantities in the array anmq.

  @param anmq - array with flags for used material types
                anmq[i] = 1 => qunatity type nonmechquant(i+1) is required
                anmq[i] = 0 => qunatity type nonmechquant(i+1) is not required

  @return The function does not return anything, but it may change content of anmq array.
*/
void camclaycoup::give_reqnmq(long *anmq)
{
  anmq[saturation_degree-1] = 1;
  anmq[suction-1] = 1;
}



/**
  The function changes material parameters for stochastic analysis.
   
  @param atm - selected material parameters (parameters which are changed)
  @param val - array containing new values of parameters
   
  @return The function does not return anything.

  Created by Tomas Koudelka,  09.2012
*/
void camclaycoup::changeparam (atsel &atm,vector &val)
{
  long i;
  
  for (i=0;i<atm.num;i++)
  {
    switch (atm.atrib[i])
    {
      case 0:
        m=val[i];
        break;
      case 1:
        lambda=val[i];
        break;
      case 2:
        kappa=val[i];
        break;
      case 3:
        c1_tilde=val[i];
        break;
      case 4:
        c2=val[i];
        break;
      case 5:
        ks=val[i];
        break;
      default:
        print_err("wrong number of atribute", __FILE__, __LINE__, __func__);
    }
  }
}
