#include "fixortodam.h"
#include "global.h"
#include "sequent.h"
#include "intpoints.h"
#include "matrix.h"
#include "vecttens.h"


fixortodam::fixortodam (void)
{
  x1t = x2t = x3t = 0.0;
  allocv(3, xeqtf);
  allocv(3, xeqt0);
}



fixortodam::~fixortodam (void)
{
}



void fixortodam::read (XFILE *in)
{
  xfscanf(in, "%k %le", "x1t", &x1t);
  xfscanf(in, "%k %le", "x2t", &x2t);
  xfscanf(in, "%k %le", "x3t", &x3t);

  xfscanf(in, "%k %le", "xeq1tf", &xeqtf(0));
  xfscanf(in, "%k %le", "xeq2tf", &xeqtf(1));
  xfscanf(in, "%k %le", "xeq3tf", &xeqtf(2));

  xfscanf(in, "%k", "loc_mat");

  t11.read(in);
  t12.read(in);
  t13.read(in);
  t21.read(in);
  t22.read(in);
  t23.read(in);
  t31.read(in);
  t32.read(in);
  t33.read(in);
}


/**
   This function initializes material model data

   @param ipp - integration point number
   @param ido - index of internal variables for given material in the ipp other array

   12/06/2012 TKo
*/
void fixortodam::initval(long ipp, long ido)
{
  long n = Mm->ip[ipp].ncompstr;
  matrix d(n,n);

  Mm->elmatstiff (d,ipp);

  xeqt0(0) = x1t/d(0,0);
  xeqt0(1) = x2t/d(1,1);
  xeqt0(2) = x3t/d(2,2);
}



void fixortodam::matstiff (matrix &d,long ipp,long ido)
{
  switch (Mp->nlman->stmat)
  {
    case initial_stiff:
      Mm->elmatstiff (d,ipp);
      break;
    case tangent_stiff:
      tmatstiff(d, ipp, ido);
      break;
    default:
      print_err("unknown type of stifness matrix is required.", __FILE__, __LINE__, __func__);
  }
}



void fixortodam::tmatstiff (matrix &d,long ipp,long ido)
{
  long i;
  vector omega(3);

  // attained values of damage parameters from the last equilibrium state
  for (i=0; i<3; i++)
    omega(i) = Mm->ip[ipp].eqother[ido+i];

  secstiffmat(d, ipp, omega);
}



void fixortodam::elmatstiff (matrix &d,long ipp)
{
  vector omegao(3);
  matrix dl(d.m, d.n);

  // It is supposed that elastic stiffness matrix is taken from the elastic orthotropic material 
  // which is given in the local coord. system of the material
  Mm->elmatstiff (dl,ipp);

  // transformation of the elastic stiffnes matrix to the global coord. system
  locmat_glob_mstiff(ipp, dl, d);
}



/** 
  Transforms %vector given in the global coordinate system to the material local coord. system.

  @param epsg - %vector of strains in the global coord. system
  @param eps - %vector of strains in the local material coord. system (output parameter)
  @param epst - tensor representation of eps (output param) 

  @return The function does not returns anything, results are passed in the output arguments.
*/
void fixortodam::glob_matcoord(long ipp, vector &epsg, vector &eps, matrix &epst)
{
  vector e1(3);
  vector e2(3);
  vector e3(3);
  matrix t(6,6);
  vector p(3);
  char *namevar[] = {"x", "y", "z"};

  // x,y,z coordinates of the integration point will be stored in the vector p
  ipcoord(Mm->elip[ipp], ipp, 0, 0, p);

  // vector of local X axis given in the global coordinate system (i.e. the first base vector of the local coord. system)
  e1(0) = t11.getval(p, namevar);
  e1(1) = t12.getval(p, namevar);
  e1(2) = t13.getval(p, namevar);

  // vector of local Y axis given in the global coordinate system (i.e. the second base vector of the local coord. system)
  e2(0) = t21.getval(p, namevar);
  e2(1) = t22.getval(p, namevar);
  e2(2) = t23.getval(p, namevar);

  // vector of local Z axis given in the global coordinate system (i.e. the third base vector of the local coord. system)
  e3(0) = t31.getval(p, namevar);
  e3(1) = t32.getval(p, namevar);
  e3(2) = t33.getval(p, namevar);

  // the first row of transformation matrix
  t(0,0) = e1(0)*e1(0);
  t(0,1) = e1(1)*e1(1);
  t(0,2) = e1(2)*e1(2);
  t(0,3) = e1(1)*e1(2);
  t(0,4) = e1(0)*e1(2);
  t(0,5) = e1(0)*e1(1);

  // the second row of transformation matrix
  t(1,0) = e2(0)*e2(0);
  t(1,1) = e2(1)*e2(1);
  t(1,2) = e2(2)*e2(2);
  t(1,3) = e2(1)*e2(2);
  t(1,4) = e2(0)*e2(2);
  t(1,5) = e2(0)*e2(1);

  // the third row of transformation matrix
  t(2,0) = e3(0)*e3(0);
  t(2,1) = e3(1)*e3(1);
  t(2,2) = e3(2)*e3(2);
  t(2,3) = e3(1)*e3(2);
  t(2,4) = e3(0)*e3(2);
  t(2,5) = e3(0)*e3(1);

  // the fourth row of transformation matrix
  t(3,0) = 2.0*e2(0)*e3(0);
  t(3,1) = 2.0*e2(1)*e3(1);
  t(3,2) = 2.0*e2(2)*e3(2);
  t(3,3) = e2(1)*e3(2)+e2(2)*e3(1);
  t(3,4) = e2(0)*e3(2)+e2(2)*e3(0);
  t(3,5) = e2(0)*e3(1)+e2(1)*e3(0);

  // the fifth row of transformation matrix
  t(4,0) = 2.0*e1(0)*e3(0);
  t(4,1) = 2.0*e1(1)*e3(1);
  t(4,2) = 2.0*e1(2)*e3(2);
  t(4,3) = e1(1)*e3(2)+e1(2)*e3(1);
  t(4,4) = e1(0)*e3(2)+e1(2)*e3(0);
  t(4,5) = e1(0)*e3(1)+e1(1)*e3(0);

  // the sixth row of transformation matrix
  t(5,0) = 2.0*e1(0)*e2(0);
  t(5,1) = 2.0*e1(1)*e2(1);
  t(5,2) = 2.0*e1(2)*e2(2);
  t(5,3) = e1(1)*e2(2)+e1(2)*e2(1);
  t(5,4) = e1(0)*e2(2)+e1(2)*e2(0);
  t(5,5) = e1(0)*e2(1)+e1(1)*e2(0);

  // eps_l = T eps_g
  mxv(t, epsg, eps);
 
  // transformation of eps to tensor notation epst
  vector_tensor (eps,epst,Mm->ip[ipp].ssst,strain);
}


/** 
  Transforms %vector given in the the material local coord. system to the global coord. system.

  @param sig - %vector of stresses in the local material coord. system (output parameter)
  @param sig_g -  %vector of stresses in the global coord. system

  @return The function does not returns anything, results are passed in the output arguments.
*/
void fixortodam::matcoord_glob(long ipp, vector &sig, vector &sig_g)
{
  vector e1(3);
  vector e2(3);
  vector e3(3);
  matrix t(6,6);
  vector p(3);
  char *namevar[] = {"x", "y", "z"};

  ipcoord(Mm->elip[ipp], ipp, 0, 0, p);

  e1(0) = t11.getval(p, namevar);
  e1(1) = t12.getval(p, namevar);
  e1(2) = t13.getval(p, namevar);

  e2(0) = t21.getval(p, namevar);
  e2(1) = t22.getval(p, namevar);
  e2(2) = t23.getval(p, namevar);

  e3(0) = t31.getval(p, namevar);
  e3(1) = t32.getval(p, namevar);
  e3(2) = t33.getval(p, namevar);

    // the first row of transp. matrix
  t(0,0) = e1(0)*e1(0);
  t(0,1) = e2(0)*e2(0);
  t(0,2) = e3(0)*e3(0);
  t(0,3) = 2.0*e2(0)*e3(0);
  t(0,4) = 2.0*e1(0)*e3(0);
  t(0,5) = 2.0*e1(0)*e2(0);

  // the second row of transp. matrix
  t(1,0) = e1(1)*e1(1);
  t(1,1) = e2(1)*e2(1);
  t(1,2) = e3(1)*e3(1);
  t(1,3) = 2.0*e2(1)*e3(1);
  t(1,4) = 2.0*e1(1)*e3(1);
  t(1,5) = 2.0*e1(1)*e2(1);

  // the third row of transp. matrix
  t(2,0) = e1(2)*e1(2);
  t(2,1) = e2(2)*e2(2);
  t(2,2) = e3(2)*e3(2);
  t(2,3) = 2.0*e2(2)*e3(2);
  t(2,4) = 2.0*e1(2)*e3(2);
  t(2,5) = 2.0*e1(2)*e2(2);

  // the fourth row of transp. matrix
  t(3,0) = e1(1)*e1(2);
  t(3,1) = e2(1)*e2(2);
  t(3,2) = e3(1)*e3(2);
  t(3,3) = e2(1)*e3(2)+e2(2)*e3(1);
  t(3,4) = e1(1)*e3(2)+e1(2)*e3(1);
  t(3,5) = e1(1)*e2(2)+e1(2)*e2(1);

  // the fifth row of transp. matrix
  t(4,0) = e1(0)*e1(2);
  t(4,1) = e2(0)*e2(2);
  t(4,2) = e3(0)*e3(2);
  t(4,3) = e2(0)*e3(2)+e2(2)*e3(0);
  t(4,4) = e1(0)*e3(2)+e1(2)*e3(0);
  t(4,5) = e1(0)*e2(2)+e1(2)*e2(0);

  // the sixth row of transp. matrix
  t(5,0) = e1(0)*e1(1);
  t(5,1) = e2(0)*e2(1);
  t(5,2) = e3(0)*e3(1);
  t(5,3) = e2(0)*e3(1)+e2(1)*e3(0);
  t(5,4) = e1(0)*e3(1)+e1(1)*e3(0);
  t(5,5) = e1(0)*e2(1)+e1(1)*e2(0);

  // sig_g = T^T sig 
  mxv(t, sig, sig_g);
}



void fixortodam::locmat_glob_mstiff(long ipp, matrix &dl, matrix &dg)
{
  vector e1(3);
  vector e2(3);
  vector e3(3);
  matrix t(6,6);
  matrix tt(6,6);
  matrix aux(6,6);
  vector p(3);
  char *namevar[] = {"x", "y", "z"};

  ipcoord(Mm->elip[ipp], ipp, 0, 0, p);

  e1(0) = t11.getval(p, namevar);
  e1(1) = t12.getval(p, namevar);
  e1(2) = t13.getval(p, namevar);

  e2(0) = t21.getval(p, namevar);
  e2(1) = t22.getval(p, namevar);
  e2(2) = t23.getval(p, namevar);

  e3(0) = t31.getval(p, namevar);
  e3(1) = t32.getval(p, namevar);
  e3(2) = t33.getval(p, namevar);

  t(0,0) = e1(0)*e1(0);
  t(0,1) = e1(1)*e1(1);
  t(0,2) = e1(2)*e1(2);
  t(0,3) = e1(1)*e1(2);
  t(0,4) = e1(0)*e1(2);
  t(0,5) = e1(0)*e1(1);

  t(1,0) = e2(0)*e2(0);
  t(1,1) = e2(1)*e2(1);
  t(1,2) = e2(2)*e2(2);
  t(1,3) = e2(1)*e2(2);
  t(1,4) = e2(0)*e2(2);
  t(1,5) = e2(0)*e2(1);

  t(2,0) = e3(0)*e3(0);
  t(2,1) = e3(1)*e3(1);
  t(2,2) = e3(2)*e3(2);
  t(2,3) = e3(1)*e3(2);
  t(2,4) = e3(0)*e3(2);
  t(2,5) = e3(0)*e3(1);

  t(3,0) = 2.0*e2(0)*e3(0);
  t(3,1) = 2.0*e2(1)*e3(1);
  t(3,2) = 2.0*e2(2)*e3(2);
  t(3,3) = e2(1)*e3(2)+e2(2)*e3(1);
  t(3,4) = e2(0)*e3(2)+e2(2)*e3(0);
  t(3,5) = e2(0)*e3(1)+e2(1)*e3(0);

  t(4,0) = 2.0*e1(0)*e3(0);
  t(4,1) = 2.0*e1(1)*e3(1);
  t(4,2) = 2.0*e1(2)*e3(2);
  t(4,3) = e1(1)*e3(2)+e1(2)*e3(1);
  t(4,4) = e1(0)*e3(2)+e1(2)*e3(0);
  t(4,5) = e1(0)*e3(1)+e1(1)*e3(0);

  t(5,0) = 2.0*e1(0)*e2(0);
  t(5,1) = 2.0*e1(1)*e2(1);
  t(5,2) = 2.0*e1(2)*e2(2);
  t(5,3) = e1(1)*e2(2)+e1(2)*e2(1);
  t(5,4) = e1(0)*e2(2)+e1(2)*e2(0);
  t(5,5) = e1(0)*e2(1)+e1(1)*e2(0);
  
  // transformation of the stiffness tensor to the global coordinate system: Dg = T^T D_l T
  // aux = D_l T
  mxm (dl,t,aux);
  // create transposed transformation matrix tt
  tranm(t,tt);
  // D_g = T^T aux 
  mxm (tt,aux,dg);
}



/**
  The function returns actual values of equivalent displacements
  in the xeq.

  @param epst - strain tensor
  @param xeq  - equivalent displacement %vector (output parameter)

*/
void fixortodam::compute_eqdispl(long ipp, matrix &epst, vector &xeq)
{   
  long i, l;
  long n = Mm->ip[ipp].ncompstr;
  matrix d(n,n);
  vector eps(n);
  vector sig(n);
  matrix sigt(n,n);

  Mm->elmatstiff (d,ipp);

  long eid = Mm->elip[ipp];
  switch (Mt->give_dimension(eid))
  {
    case 1:
      l = Mt->give_length(eid);
      break;
    case 2:
      l = sqrt(Mt->give_area(eid));
      break;
    case 3:
      l = pow(Mt->give_volume(eid), 1.0/3.0);
      break;
    default:
      print_err("unknown dimension of element", __FILE__, __LINE__, __func__);
  }

  tensor_vector(eps,epst,Mm->ip[ipp].ssst,strain);
  mxv(d,eps,sig);
  vector_tensor(sig,sigt,Mm->ip[ipp].ssst,stress);

  for (i=0; i<3; i++)
  {
    xeq(i) = sigt(i,i)/d(i,i) * l;
    if (xeq(i) < 0.0)
      xeq(i) = 0.0;
  }
  
  
  
/*  for (i=0; i<3; i++)
    xeq(i)=l*sqrt(pow(epst(i,i),2));  */
}



/**
  The function returns actual values of damage parameters in the particular
  loacl material directions stored in the %vector omega

  @param xeq  - equivalent displacement %vector
  @param omegao - %vector of damage parametrs from the last equilibrium state
  @param omega - %vector of calculated damage parametrs (output parameter)

*/
void fixortodam::compute_dam (vector &xeq, vector &omegao, vector &omega)
{ 
  long i;

  for (i=0; i<3; i++)
  {
    if (xeq(i) >= xeqtf(i))
    {
      omega(i) = 1.0;
      continue;
    }

    if (xeq(i) <= xeqt0(i))
    {
      omega(i) = 0.0;
      continue;
    }
    omega(i)=xeqtf(i)*(xeq(i)-xeqt0(i))/(xeq(i)*(xeqtf(i)-xeqt0(i)));
    if (omega(i)<omegao(i))
      omega(i)=omegao(i);
    if (omega(i)<0)
      omega(i)=0;
  }
}



/**
  The function returns actual values of damage parameters in the particular
  local material directions stored in the %vector omega


  @param d - secant stiffness %matrix  (output parameter)
  @param omega - %vector of damage parameters 
  @param ipp - integration point id

*/
void fixortodam::secstiffmat(matrix &d, long ipp, vector &omega)
{ 

  // It is supposed that elastic stiffness matrix is taken from the elastic orthotropic material 
  // which is given in the local coord. system of the material
  Mm->elmatstiff (d,ipp);
  
  d(0,0)=pow(1-omega(0),2)*d(0,0);
  d(0,1)=(1-omega(0))*(1-omega(1))*d(0,1);
  d(0,2)=(1-omega(0))*(1-omega(2))*d(0,2);
  d(1,0)=d(0,1);
  d(1,1)=pow(1-omega(1),2)*d(1,1);
  d(1,2)=(1-omega(1))*(1-omega(2))*d(1,2);
  d(2,0)=d(0,2);
  d(2,1)=d(1,2);
  d(2,2)=pow(1-omega(2),2)*d(2,2);
  d(3,3)=pow((2*(1-omega(0))*(1-omega(1)))/(2-omega(0)-omega(1)),2);
  d(4,4)=pow((2*(1-omega(0))*(1-omega(2)))/(2-omega(0)-omega(2)),2);
  d(5,5)=pow((2*(1-omega(1))*(1-omega(2)))/(2-omega(1)-omega(2)),2);
}



void fixortodam::nlstresses (long ipp, long im, long ido)
{
  long i;
  long ncompstr=Mm->ip[ipp].ncompstr; // number of stress/strain components in vector form

  vector omega(3), omegao(3), xeq(3);
  vector epsg(ncompstr), eps(ncompstr), sig(ncompstr), sig_g(ncompstr);
  matrix epst(3,3), sigt(3,3), d(ncompstr, ncompstr);

  // actual strains in the global coord. system
  for (i=0; i<ncompstr; i++)
    epsg(i) = Mm->ip[ipp].strain[i];

  // transform strain vector to local coord. system of the material
  glob_matcoord(ipp, epsg, eps, epst);

  // eps contains vector representation of strains in local coord. system of the material
  // epst contains tensor representation of strains in local coord. system of the material

  // attained values of damage parameters from the last equilibrium state
  for (i=0; i<3; i++)
    omegao(i) = Mm->ip[ipp].eqother[ido+i];

  // compute equivalent displacements
  compute_eqdispl(ipp, epst, xeq);

  // compute actual damage parameters
  compute_dam(xeq, omegao, omega);

  // secant stiffnes matrix
  secstiffmat(d, ipp, omega);
  
  // compute actual stress vector in the local coord. system of the material
  // sig = D(omega) * eps
  mxv(d, eps, sig);
  
  // transforms stress vector in the local coord. system of the material to global coord. system
  matcoord_glob(ipp, sig, sig_g);

  // store actual stress in the global coord. system to the integration point
  for (i=0; i<ncompstr; i++)
    Mm->ip[ipp].stress[i] = sig_g[i];

  
  // store actual damage parameters to the integration point
  for (i=0; i<3; i++)
    Mm->ip[ipp].other[ido+i] = omega(i);
}



void fixortodam::updateval (long ipp,long im,long ido)
{
  long i;

  for (i=0; i<3; i++)
    Mm->ip[ipp].eqother[ido+i] = Mm->ip[ipp].other[ido+i];
}
