#include <stdlib.h>
#include <string.h>

#include "mechbclc.h"
#include "global.h"
#include "alias.h"
#include "loadcase.h"
#include "dloadcase.h"
#include "inicd.h"
#include "vector.h"
#include "matrix.h"
#include "elemhead.h"
#include "gtopology.h"
#include "element.h"
#include "node.h"
#include "elemswitch.h"
#include "intpoints.h"


/**
  Constructor initializes data members to zero or default values.

  Created by JK,
  Modified by TKo, TKr,
*/
mechbclc::mechbclc ()
{
  //  number of load cases
  nlc=0L;
  //  number of initial conditions
  nico = 0L;
  //  load case
  lc=NULL;
  //  time dependent load case
  dlc = NULL;
  ico = NULL;
  ncsum = 0L;
  sumcomp         = NULL;
  reactsumcomp    = NULL;
  pd_reactsumcomp = NULL;

  //  the number of general functions describing the eigenstrains
  ngfes=0L;
  //  array of general functions describing the eigenstrains
  eigstrfun = NULL;
  strastre = (strastrestate)0;
}



/**
  Destructor releases allocated memory of the mechbclc object.

  Created by JK,
  Modified by TKo, TKr,
*/
mechbclc::~mechbclc ()
{
  if (lc!=NULL)
    delete [] lc;
  if (dlc!=NULL)
    delete [] dlc;
  if (ico!=NULL)
    delete [] ico;
  if (sumcomp!=NULL)
    delete [] sumcomp;
  if (reactsumcomp!=NULL)
    delete [] reactsumcomp;
  if (pd_reactsumcomp!=NULL)
    delete [] pd_reactsumcomp;
  if (eigstrfun!=NULL)
    delete [] eigstrfun;
}



/**
  Function reads data about boundary conditions and load cases.
   
  @param in - pointer to the opened XFILE
   
  @return The function does not return anything.

  Created by JK,
  Modified by Tomas Koudelka,
*/
void mechbclc::read (XFILE *in)
{
  long i;
  vector nodval,ipval;
  
  switch (Mp->tprob){
  case linear_statics:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc<0){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
    }
    break;
  }
  case mat_nonlinear_statics:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc!=1){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [2*nlc];
    for (i=0;i<2*nlc;i++){
      lc[i].read (in);
    }
    readinic(in);
    break;
  }
  case geom_nonlinear_statics:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc!=1){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [2*nlc];
    for (i=0;i<2*nlc;i++){
      lc[i].read (in);
    }
    readinic(in);
    break;
  }
  case eigen_dynamics:{
    if (Mp->straincomp==1 || Mp->stresscomp==1)
      nlc=Mp->eigsol.neigv;
    break;
  }
  case forced_dynamics:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc<0){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    dlc = new dloadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
      dlc[i].read (in);
    }
    break;
  }
  case mech_timedependent_prob:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc!=1){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    dlc = new dloadcase [nlc];
    for (i=0;i<nlc;i++){
      dlc[i].read (in);
    }
    readinic(in);
    break;
  }
  case growing_mech_structure:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc!=1){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    dlc = new dloadcase [nlc];
    for (i=0;i<nlc;i++){
      dlc[i].read (in);
    }
    readinic(in);
    break;
  }
  case earth_pressure:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (Mp->tepsol < epplast_sol)
    {
      lc = new loadcase [nlc];
      dlc = new dloadcase [nlc];
      for (i=0;i<nlc;i++){
        lc[i].read (in);
        dlc[i].read (in);
      }
      readinic(in);
    }
    else
    {
      lc = new loadcase [2*nlc];
      for (i=0;i<2*nlc;i++){
        lc[i].read (in);
      }
      readinic(in);
    }
    break;
  }
  case layered_linear_statics:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc<0){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
    }
    break;
  }
    
  case lin_floating_subdomain:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc<0){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
    }
    break;
  }

  case nonlin_floating_subdomain:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc!=1){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [2*nlc];
    for (i=0;i<2*nlc;i++){
      lc[i].read (in);
    }
    break;
  }

  case nonlinear_dynamics:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc!=1){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    dlc = new dloadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
      dlc[i].read (in);
    }
    break;
  }
    
  case hemivar_inequal:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc<0){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
    }
    break;
  }
    
  case load_balancing:{
    xfscanf (in,"%k%ld","number_of_load_cases",&nlc);
    if (nlc<0){
      print_err("wrong number of load cases", __FILE__, __LINE__, __func__);
      abort ();
    }
    lc = new loadcase [nlc];
    for (i=0;i<nlc;i++){
      lc[i].read (in);
    }
    break;
  }

  default:{
    print_err("unknown problem type", __FILE__, __LINE__, __func__);
  }
  }
  
  // **********************
  //  eigenstrain reading
  // **********************
  read_eigenstrains (in);
  
}


/**
  Function prints data about boundary conditions and load cases.
   
  @param out - pointer to the opened FILE
   
  @return The function does not return anything.

  TKr, 07/02/2013 according to read (XFILE *in)
*/
void mechbclc::print (FILE *out)
{
  long i;
  
  fprintf (out,"\n number of load cases:");
  switch (Mp->tprob){
  case linear_statics:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
    }
    break;
  }
  case mat_nonlinear_statics:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<2*nlc;i++){
      lc[i].print (out);
    }
    printinic(out);
    break;
  }
  case geom_nonlinear_statics:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<2*nlc;i++){
      lc[i].print (out);
    }
    printinic(out);
    break;
  }
  case eigen_dynamics:{
    break;
  }
  case forced_dynamics:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
      dlc[i].print (out);
    }
    break;
  }
  case mech_timedependent_prob:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      dlc[i].print (out);
    }
    printinic(out);
    break;
  }
  case growing_mech_structure:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      dlc[i].print (out);
    }
    printinic(out);
    break;
  }
  case earth_pressure:{
    fprintf (out,"\n  %ld",nlc);
    if (Mp->tepsol < epplast_sol)
    {
      for (i=0;i<nlc;i++){
        lc[i].print (out);
        dlc[i].print (out);
      }
      printinic(out);
    }
    else
    {
      for (i=0;i<2*nlc;i++){
        lc[i].print (out);
      }
      printinic(out);
    }
    break;
  }
  case layered_linear_statics:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
    }
    break;
  }
    
  case lin_floating_subdomain:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
    }
    break;
  }

  case nonlin_floating_subdomain:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<2*nlc;i++){
      lc[i].print (out);
    }
    break;
  }

  case nonlinear_dynamics:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
      dlc[i].print (out);
    }
    break;
  }
    
  case hemivar_inequal:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
    }
    break;
  }
    
  case load_balancing:{
    fprintf (out,"\n  %ld",nlc);
    for (i=0;i<nlc;i++){
      lc[i].print (out);
    }
    break;
  }

  default:{
    print_err("unknown problem type", __FILE__, __LINE__, __func__);
  }
  }
  
  // **********************
  //  eigenstrain printing
  // **********************
  print_eigenstrains (out);
  
}


/**
   function reads eigenstrains
   
   @param in - input file
   
   JK, 8.11.2012
*/
void mechbclc::read_eigenstrains (XFILE *in)
{
  long i,j,k;
  long ncomp,nip,ipid;
  long *id;
  
  //  Mp->eigstrains = 0 - eigenstrains are not defined
  //  Mp->eigstrains = 1 - eigenstrains are defined for the whole problem
  //  Mp->eigstrains = 2 - eigenstrains are defined on each element
  //  Mp->eigstrains = 3 - eigenstrains are defined outside of MEFEL (for the coupled problems)
  //  Mp->eigstrains = 4 - eigenstresses are defined for the whole problem
  //  Mp->eigstrains = 5 - eigenstresses are defined on each element
  xfscanf (in,"%k%ld","eigenstrains",&(Mp->eigstrains));
  if (Mp->eigstrains<0 || Mp->eigstrains>5){
    print_err("wrong definition of eigenstrains", __FILE__, __LINE__, __func__);
    abort ();
  }
  
  if (Mespr == 1){
    if (Mp->eigstrains==0)
      fprintf(stdout, "\n eigenstrains are not defined");
    if (Mp->eigstrains==1)
      fprintf(stdout, "\n eigenstrains are defined for the whole problem by a single set of values");
    if (Mp->eigstrains==2)
      fprintf(stdout, "\n eigenstrains are defined independently for each element");
    if (Mp->eigstrains==3)
      fprintf(stdout, "\n eigenstrains are defined independently in TRFEL");
    if (Mp->eigstrains==4)
      fprintf(stdout, "\n eigenstresses are defined for the whole problem by a single set of values");
    if (Mp->eigstrains==5)
      fprintf(stdout, "\n eigenstresses are defined independently for each element");
  }
  
  
  if (Mp->eigstrains>0){
    //  eigenstrains are defined
    
    //  array of eigenstrain id
    if (Mm->eigstrid==NULL)
      Mm->eigstrid = new long* [Mm->tnip];
    //  array of eigenstrains in integration points
    if (Mm->eigstrains==NULL)
      Mm->eigstrains = new double* [Mm->tnip];
    //  array of eigenstresses in integration points
    if (Mm->eigstresses==NULL)
      Mm->eigstresses = new double* [Mm->tnip];
    
    id = new long [6];
  }
  
  
  // *************************************************
  //  eigenstrains are defined for the whole problem
  // *************************************************
  if (Mp->eigstrains==1){
    //  type of strain/stress state
    //  strastre = 1 - bar
    //  strastre = 10 - plane stress
    //  strastre = 30 - space strain/stress
    xfscanf (in,"%m", &strastrestate_kwdset, (int *)&strastre);
    
    switch (strastre){
    case bar:{
      xfscanf (in,"%ld",id);
      ncomp=1;
      break;
    }
    case planestress:{
      xfscanf (in,"%ld %ld %ld",id,id+1,id+2);
      ncomp=3;
      break;
    }
    case spacestress:{
      xfscanf (in,"%ld %ld %ld %ld %ld %ld",id,id+1,id+2,id+3,id+4,id+5);
      ncomp=6;
      break;
    }
    default:{
      print_err("unknown type of strain-stress state", __FILE__, __LINE__, __func__);
    }
    }//  end of the switch (strastre){
    
    for (i=0;i<Mm->tnip;i++){
      //  array of eigenstrains in integration points
      Mm->eigstrains[i] = new double [ncomp];
      memset(Mm->eigstrains[i], 0, sizeof(*Mm->eigstrains[i])*ncomp);
      //  array of eigenstresses in integration points
      Mm->eigstresses[i] = new double [ncomp];
      memset(Mm->eigstresses[i], 0, sizeof(*Mm->eigstresses[i])*ncomp);
      //  array of eigenstrain id
      Mm->eigstrid[i] = new long [ncomp];
      
      for (j=0;j<ncomp;j++){
	Mm->eigstrid[i][j]=id[j]-1;
      }
    }
    
  }//  end of the statement if (Mp->eigstrains==1){
  
  
  // ***************************************
  //  eigenstrains are defined on elements
  // ***************************************
  if (Mp->eigstrains==2){
    
    //  loop over the number of elements
    for (i=0;i<Mt->ne;i++){
      //  strain/stress state on element
      strastre = Mt->give_ssst (i,0);
      
      switch (strastre){
      case bar:{
	xfscanf (in,"%ld",id);
	ncomp=1;
	break;
      }
      case planestress:{
	xfscanf (in,"%ld %ld %ld",id,id+1,id+2);
	ncomp=3;
	break;
      }
      case spacestress:{
	xfscanf (in,"%ld %ld %ld %ld %ld %ld",id,id+1,id+2,id+3,id+4,id+5);
	ncomp=6;
	break;
      }
      default:{
	print_err("unknown type of strain-stress state", __FILE__, __LINE__, __func__);
      }
      }//  end of the switch (strastre){
      
      //  number of integration points on element
      nip = Mt->give_tnip (i);
      //  number of the first integration point
      ipid=Mt->elements[i].ipp[0][0];
      
      //  loop over the number of integration points on element
      for (j=0;j<nip;j++){
	Mm->eigstrains[ipid] = new double [ncomp];
	Mm->eigstresses[ipid] = new double [ncomp];
	Mm->eigstrid[ipid] = new long [ncomp];
	
	for (k=0;k<ncomp;k++){
	  Mm->eigstrid[ipid][k] = id[k]-1;
	}
	
	ipid++;
      }//  end of the loop over the number of integration points on element
      
    }//  end of the loop over the number of elements 
  }//  end of the if (Mp->eigstrains==2)
  
  
  if (Mp->eigstrains==3){
    //  loop over the number of integration points in problem
    for (j=0;j<Mm->tnip;j++){
      ncomp = Mm->ip[j].ncompstr;
      Mm->eigstrains[j] = new double[ncomp];
      memset(Mm->eigstrains[j], 0, sizeof(*Mm->eigstrains[j])*ncomp);
      Mm->eigstresses[j] = new double [ncomp];
      memset(Mm->eigstresses[j], 0, sizeof(*Mm->eigstresses[j])*ncomp);
    }//  end of the loop over the number of integration points on element
    
  }//  end of the statement if (Mp->eigstrains==3){
  
  
  // ***************************************
  //  eigenstresses are defined on elements
  // ***************************************
  if (Mp->eigstrains==4){
    //  type of strain/stress state
    //  strastre = 1 - bar
    //  strastre = 10 - plane stress
    //  strastre = 30 - space strain/stress
    xfscanf (in,"%m", &strastrestate_kwdset, (int *)&strastre);
    
    switch (strastre){
    case bar:{
      xfscanf (in,"%ld",id);
      ncomp=1;
      break;
    }
    case planestress:{
      xfscanf (in,"%ld %ld %ld",id,id+1,id+2);
      ncomp=3;
      break;
    }
    case axisymm:{
      xfscanf(in,"%ld %ld %ld %ld", id, id+1, id+2, id+3);
      ncomp = 4;
      break;
    }
    case spacestress:{
      xfscanf (in,"%ld %ld %ld %ld %ld %ld",id,id+1,id+2,id+3,id+4,id+5);
      ncomp=6;
      break;
    }
    default:{
      print_err("unknown type of strain-stress state", __FILE__, __LINE__, __func__);
    }
    }//  end of the switch (strastre){
    
    for (i=0;i<Mm->tnip;i++){
      //  array of eigenstrains in integration points
      Mm->eigstrains[i] = new double [ncomp];
      memset(Mm->eigstrains[i], 0, sizeof(*Mm->eigstrains[i])*ncomp);
      //  array of eigenstresses in integration points
      Mm->eigstresses[i] = new double [ncomp];
      memset(Mm->eigstresses[i], 0, sizeof(*Mm->eigstresses[i])*ncomp);
      //  array of eigenstrain id
      Mm->eigstrid[i] = new long [ncomp];
      
      for (j=0;j<ncomp;j++){
	Mm->eigstrid[i][j]=id[j]-1;
      }
    }
    
  }//  end of the statement if (Mp->eigstrains==1){
  
  
  // ***************************************
  //  eigenstresses are defined on elements
  // ***************************************
  if (Mp->eigstrains==5){
    
    //  loop over the number of elements
    for (i=0;i<Mt->ne;i++){
      //  strain/stress state on element
      strastre = Mt->give_ssst (i,0);
      
      switch (strastre){
      case bar:{
	xfscanf (in,"%ld",id);
	ncomp=1;
	break;
      }
      case planestress:{
	xfscanf (in,"%ld %ld %ld",id,id+1,id+2);
	ncomp=3;
	break;
      }
      case axisymm:{
        xfscanf(in,"%ld %ld %ld %ld", id, id+1, id+2, id+3);
        ncomp = 4;
        break;
      }
      case spacestress:{
	xfscanf (in,"%ld %ld %ld %ld %ld %ld",id,id+1,id+2,id+3,id+4,id+5);
	ncomp=6;
	break;
      }
      default:{
	print_err("unknown type of strain-stress state", __FILE__, __LINE__, __func__);
      }
      }//  end of the switch (strastre){
      
      //  number of integration points on element
      nip = Mt->give_tnip (i);
      //  number of the first integration point
      ipid=Mt->elements[i].ipp[0][0];
      
      //  loop over the number of integration points on element
      for (j=0;j<nip;j++){
      //  array of eigenstrains in integration points
        Mm->eigstrains[ipid] = new double [ncomp];
        memset(Mm->eigstrains[ipid], 0, sizeof(*Mm->eigstrains[ipid])*ncomp);
	Mm->eigstresses[ipid] = new double [ncomp];
	Mm->eigstrid[ipid] = new long [ncomp];
	
	for (k=0;k<ncomp;k++){
	  Mm->eigstrid[ipid][k] = id[k]-1;
	}
	
	ipid++;
      }//  end of the loop over the number of integration points on element
      
    }//  end of the loop over the number of elements 
  }//  end of the if (Mp->eigstrains==2)



  if (Mp->eigstrains>0){
    delete [] id;
  }
  
  // ************************************************
  //  general functions describing the eigenstrains
  // ************************************************
  if (Mp->eigstrains==1 || Mp->eigstrains==2 || Mp->eigstrains==4 || Mp->eigstrains==5){
    //  the number of general functions needed for eigenstrain description
    xfscanf (in,"%k%ld","number_of_gfunct_eigenstr",&ngfes);
    
    eigstrfun = new gfunct [ngfes];
    
    //  loop over the number of general functions
    for (i=0;i<ngfes;i++){
      eigstrfun[i].read (in);
    }//  end of the loop over the number of general functions
    
    //  computation of eigstrains
    eigstrain_computation (Mp->time);
    
  }//  end of the statement if (Mp->eigstrains==1 || Mp->eigstrains==2){

}



/**
   function prints eigenstrains, according to read_eigenstrains (XFILE *in)
   
   @param out - output file
   
   TKr, 07/02/2013
*/
void mechbclc::print_eigenstrains (FILE *out)
{
  long i;
  long ipid;
  
  //  Mp->eigstrains = 0 - eigenstrains are not defined
  //  Mp->eigstrains = 1 - eigenstrains are defined for the whole problem
  //  Mp->eigstrains = 2 - eigenstrains are defined on each element
  //  Mp->eigstrains = 3 - eigenstrains are defined outside of MEFEL (for the coupled problems)
  fprintf (out,"\n\n%ld",Mp->eigstrains);
  
  // *************************************************
  //  eigenstrains are defined for the whole problem
  // *************************************************
  if ((Mp->eigstrains==1) || (Mp->eigstrains==4)){
    //  type of strain/stress state
    //  strastre = 1 - bar
    //  strastre = 10 - plane stress
    //  strastre = 30 - space strain/stress
    fprintf (out,"\n\n%d",(int)strastre);
    
    switch (strastre){
    case bar:{
      fprintf (out,"\n  %ld",Mm->eigstrid[0][0]+1);
      break;
    }
    case planestress:{
      fprintf (out,"\n  %ld %ld %ld",Mm->eigstrid[0][0]+1,Mm->eigstrid[0][1]+1,Mm->eigstrid[0][2]+1);
      break;
    }
    case axisymm:{
      fprintf (out,"\n  %ld %ld %ld %ld",Mm->eigstrid[0][0]+1,Mm->eigstrid[0][1]+1,Mm->eigstrid[0][2]+1,Mm->eigstrid[0][3]+1);
      break;
    }
    case spacestress:{
      fprintf (out,"\n  %ld %ld %ld %ld %ld %ld",Mm->eigstrid[0][0]+1,Mm->eigstrid[0][1]+1,Mm->eigstrid[0][2]+1
	       ,Mm->eigstrid[0][3]+1,Mm->eigstrid[0][4]+1,Mm->eigstrid[0][5]+1);
      break;
    }
    default:{
      print_err("unknown type of strain-stress state", __FILE__, __LINE__, __func__);
    }
    }//  end of the switch (strastre){
  }//  end of the statement if (Mp->eigstrains==1){
  
  
  // ***************************************
  //  eigenstrains are defined on elements
  // ***************************************
  if ((Mp->eigstrains==2) || (Mp->eigstrains==5)){
    
    //  loop over the number of elements
    for (i=0;i<Mt->ne;i++){
      
      //  strain/stress state on element
      strastre = Mt->give_ssst (i,0);
      
      //  number of the first integration point
      ipid=Mt->elements[i].ipp[0][0];
      
      switch (strastre){
      case bar:{
	fprintf (out,"\n  %ld",Mm->eigstrid[ipid][0]+1);
	break;
      }
      case planestress:{
	fprintf (out,"\n  %ld %ld %ld",Mm->eigstrid[ipid][0]+1,Mm->eigstrid[ipid][1]+1,Mm->eigstrid[ipid][2]+1);
	break;
      }
      case axisymm:{
	fprintf (out,"\n  %ld %ld %ld %ld",Mm->eigstrid[ipid][0]+1,Mm->eigstrid[ipid][1]+1,Mm->eigstrid[ipid][2]+1,Mm->eigstrid[ipid][3]+1);
	break;
      }
      case spacestress:{
	fprintf (out,"\n  %ld %ld %ld %ld %ld %ld",Mm->eigstrid[ipid][0]+1,Mm->eigstrid[ipid][1]+1,Mm->eigstrid[ipid][2]+1
		 ,Mm->eigstrid[ipid][3]+1,Mm->eigstrid[ipid][4]+1,Mm->eigstrid[ipid][5]+1);
	break;
      }
      default:{
	print_err("unknown type of strain-stress state", __FILE__, __LINE__, __func__);
      }
      }//  end of the switch (strastre){
    }//  end of the loop over the number of elements 
  }//  end of the if (Mp->eigstrains==2)
  

  // ************************************************
  //  general functions describing the eigenstrains
  // ************************************************
  if (Mp->eigstrains==1 || Mp->eigstrains==2 || Mp->eigstrains==4 || Mp->eigstrains==5){
    //  the number of general functions needed for eigenstrain description
    fprintf (out,"\n\n  %ld",ngfes);
    
    //  loop over the number of general functions
    for (i=0;i<ngfes;i++){
      eigstrfun[i].print (out);
    }//  end of the loop over the number of general functions
    
  }//  end of the statement if (Mp->eigstrains==1 || Mp->eigstrains==2){

}


/**
   function computes eigenstrains
   
   @param time - actual time
   
   8. 11. 2012
*/
void mechbclc::eigstrain_computation (double time)
{
  long i,j,gfid,ncomp;
  
  //  loop over the number of all integration points
  for (i=0;i<Mm->tnip;i++){
    //  the number of strain components in the integration point
    ncomp = Mm->ip[i].give_ncompstr ();
    
    //  loop over the number of eigenstrain components
    for (j=0;j<ncomp;j++){
      //  id of the general function
      gfid = Mm->eigstrid[i][j];
      
      if (Mp->eigstrains < 3)
        Mm->eigstrains[i][j] = eigstrfun[gfid].getval (time);
      if (Mp->eigstrains > 3)
        Mm->eigstresses[i][j] = -eigstrfun[gfid].getval (time);
      
    }//  end of the loop over the number of eigenstrain components
    
  }//  end of the loop over the number of all integration points
  
}

/**
  This method reads section with data about initial conditions in the nodes
  and if it is necessarry it allocates memory for mechmat ic array.
  
  @param in is pointer to the structure with opened file.

  @retval 0 - on succes
  @retval 1 - error reading number of nodes with initial condition
  @retval 2 - error reading node number

  Created by Tomas Koudelka,
*/
long mechbclc::readinic (XFILE *in)
{
  long i, n;
 
  xfscanf(in, "%k%ld", "num_ini_cond", &nico);

  if ((nico < 0) && (nico > Mt->nn))
  {
    print_err("invalid number of inital conditions", __FILE__, __LINE__, __func__);
    return 1;
  }
  if (nico == 0)
    return 0;
  ico = new inicd[Mt->nn];
  for (i = 0L; i < nico; i++)
  {
    xfscanf(in, "%ld", &n);
    if ((n < 0) || (n > Mt->nn))
    {
      print_err("invalid node number", __FILE__, __LINE__, __func__);
      return 2;
    }
    ico[n-1].read(in);
    if ((ico[n-1].type & inicond) && (Mm->ic == NULL))
    {
      Mm->ic = new double *[Mm->tnip];
      memset(Mm->ic, 0, sizeof(*Mm->ic)*Mm->tnip);
    }
  }
  return 0;
}



/**
  This function prints section with data about initial conditions in the nodes
  and if it is necessarry it allocates memory for mechmat ic array.
  
  @param out is pointer to the structure with opened file.

  @retval 0 - on succes
  @retval 1 - error reading number of nodes with initial condition
  @retval 2 - error reading node number

  TKr, 11/02/2013 according to readinic (XFILE *in)
*/
long mechbclc::printinic (FILE *out)
{
  long i, n=1;

  fprintf(out, "\n ## initial conditions:");
  fprintf(out, "\n %ld\n", nico);

  for (i = 0L; i < nico; i++)
  {
    fprintf(out, "\n %ld ",n);//tady vymyslet??!!
    ico[n-1].print(out);
  }
  return 0;
}



/**
  Function computes initial values at integration points
  from initial nodal values.
   
  @return The function does not return anything.

  Created by Tomas Koudelka,
*/
void mechbclc::inicipval(void)
{
  long i, j, k, l, nne, nval, eid;
  elemtype et;
  ivector enod;
  matrix  nodval;
  inictype *ictn;   // type of initial condition in the nodes


  for (i = 0; i < Mt->ne; i++)
  {
    if (Gtm->leso[i]==1)
    {      
      nne = Mt->give_nne(i);
      allocv(nne, enod);
      // find maximum number of initial values prescribed at nodes
      // initial displacement values are not taken into account
      Mt->give_elemnodes (i, enod);
      nval = ico[enod[0]].nval;
      if (ico[enod[0]].type & inidisp)
        nval -= Gtm->give_ndofn(enod[0]);
      for (j = 0; j < nne; j++)
      {
        if (nval < ico[enod[j]].nval)
        {
          nval = ico[enod[j]].nval;
          if (ico[enod[j]].type & inidisp)
            nval -= Gtm->give_ndofn(enod[j]);
        }
      }
      allocm (nne, nval, nodval);
      fillm(0.0, nodval);
      ictn = new inictype[nne];
      for (j = 0; j < nne; j++)
      {
        ictn[j] = ico[enod[j]].type;
        for (k = 0; k < nval; k++)
        {
          if (ictn[j] & inidisp)  // skip initial nodal displacements
            l = k + Gtm->give_ndofn(enod[j]);
          else
            l = k;
          nodval[j][k] = ico[enod[j]].val[l];
        }
      }
      eid = i;
      et = Mt->give_elem_type(eid);
      switch (et)
      {
	case bar2d:{              Bar2d->inicipval(eid, 0, 0, nodval, ictn);   break; }
	case barq2d:{             Barq2d->inicipval(eid, 0, 0, nodval, ictn);  break; }
	  //      case beam2d:{             Beam2d->inicipval(eid, 0, 0, nodval, ictn);  break; }
	  //      case beam3d:{             Beam3d->inicipval(eid, 0, 0, nodval, ictn);  break; }
	  //      case subsoilbeam:{        Sbeam->inicipval(eid, 0, 0, nodval, ictn);   break; }
	case spring_1:{           Spring->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case spring_2:{           Spring->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case spring_3:{           Spring->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case spring_4:{           Spring->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case spring_5:{           Spring->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case spring_6:{           Spring->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case planeelementlt:{     Pelt->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case planeelementqt:{     Peqt->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case planeelementrotlt:{  Perlt->inicipval(eid, 0, 0, nodval, ictn);   break; }
	case planeelementlq:{     Pelq->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case planeelementqq:{     Peqq->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case planeelementrotlq:{  Perlq->inicipval(eid, 0, 0, nodval, ictn);   break; }
	case planeelementsubqt:{  Pesqt->inicipval(eid, 0, 0, nodval, ictn);   break; }
	case cctel:{              Cct->inicipval(eid, 0, 0, nodval, ictn);     break; }
	case dktel:{              Dkt->inicipval(eid, 0, 0, nodval, ictn);     break; }
	case dstel:{              Dst->inicipval(eid, 0, 0, nodval, ictn);     break; }
	case q4plateel:{          Q4pl->inicipval(eid, 0, 0, nodval, ictn);    break; }
	  //      case subsoilplatetr:{     Spltr->inicipval(eid, 0, 0, nodval, ictn);   break; }
	  //      case subsoilplateq:{      Splq->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case axisymmlq:{          Asymlq->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case axisymmlt:{          Asymlt->inicipval(eid, 0, 0, nodval, ictn);  break; }
	case shelltrelem:{        Shtr->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case shellqelem:{         Shq->inicipval(eid, 0, 0, nodval, ictn);     break; }
	case lineartet:{          Ltet->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case quadrtet:{           Qtet->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case linearhex:{          Lhex->inicipval(eid, 0, 0, nodval, ictn);    break; }
	case quadrhex:{           Qhex->inicipval(eid, 0, 0, nodval, ictn);    break; }
	default:
        {
          print_err("unknown element type is required", __FILE__, __LINE__, __func__);
        }
      }
      destrv(enod);
      destrm(nodval);
      delete [] ictn;
    }
  }
}



/**
  Function allocates array containing sums of components 
  in particular directions of the nodal force %vector.
   
  @return The function does not return anything.

  Created by JK, 6.10.2003
*/
void mechbclc::alloc_sumcomp ()
{
  ncsum = Mt->give_ndofn (0);
  sumcomp = new double [ncsum];
  reactsumcomp = new double [ncsum];
  pd_reactsumcomp = new double [ncsum];
  memset(sumcomp, 0, sizeof(*sumcomp)*ncsum);
  memset(reactsumcomp, 0, sizeof(*reactsumcomp)*ncsum);
  memset(pd_reactsumcomp, 0, sizeof(*pd_reactsumcomp)*ncsum);
}



/**
  Function computes sums of components in particular directions 
  of the nodal force %vector.
   
  @param rhs - %vector of right hand side (nodal forces)
  
  Created by JK, 6.10.2003
*/
void mechbclc::comp_sum (double *rhs)
{
  long i,j,ndofn,cn;
  long *aux;
  
  aux=new long [Ndofm];
  for (i=0;i<Ndofm;i++){
    aux[i]=0;
  }
  
  nullv (sumcomp,ncsum);
  
  for (i=0;i<Mt->nn;i++){
    ndofn=Gtm->gnodes[i].ndofn;
    for (j=0;j<ndofn;j++){
      cn=Mt->give_dof (i,j);
      if (cn>0){
	if (aux[cn-1]==0){
	  sumcomp[j]+=rhs[cn-1];
	  aux[cn-1]=1;
	}
      }
    }
  }

  delete [] aux;
}



/**
  The function computes sums of components in particular directions of the reactions at supports.
  Particular sums of components are stored in reactsumcomp array.

  @return The function does not return anything.

  Created by Tomas Koudelka,  12.2009
*/
void mechbclc::comp_sum_react()
{
  long i,j,ndofn,cn;
  
  nullv (reactsumcomp,ncsum);
  
  for (i=0;i<Mt->nn;i++){
    ndofn=Gtm->gnodes[i].ndofn;
    for (j=0;j<ndofn;j++){
      cn=Mt->give_dof(i,j);
      if (cn==0)
       reactsumcomp[j]+=Mt->nodes[i].r[j];
    }
  }
}



/**
  The function computes sums of components in particular directions of the reactions caused by prescribed displacements.
  Particular sums of components are stored in pd_reactsumcomp array.
   
  @return The function does not return anything.

  Created by Tomas Koudelka,  12.2009
*/
void mechbclc::comp_sum_pdreact()
{
  long i,j,ndofn,cn;
  
  nullv (sumcomp,ncsum);
  
  for (i=0;i<Mt->nn;i++){
    ndofn=Gtm->gnodes[i].ndofn;
    for (j=0;j<ndofn;j++){
      cn=Mt->give_dof (i,j);
      if (cn<0)
        pd_reactsumcomp[j]+=Mt->nodes[i].r[j];
    }
  }
}



/**
  Function returns sums of components in particular directions of the nodal force %vector.
   
  Created by JK, 6.10.2003
*/
void mechbclc::give_comp_sum (double *sum)
{
  long i;
  for (i=0;i<ncsum;i++)
    sum[i]=sumcomp[i];
}

