#include "plquadcontact.h"
#include "global.h"
#include "element.h"
#include "node.h"
#include "globmat.h"
#include "intp.h"
#include "basefun.h"

plquadcontact::plquadcontact (void)
{
  long i;

  nne=4;  ndofe=8;
  nb=1;  tncomp=2;

  ssst = planecontact;

  ncomp = new long [nb];
  ncomp[0]=2;

  nip = new long* [nb];
  intordsm = new long* [nb];
  for (i=0;i<nb;i++){
    nip[i] = new long [nb];
    intordsm[i] = new long [nb];
  }
  
  nip[0][0]=3;
  intordsm[0][0]=3;
  tnip=3;
}



plquadcontact::~plquadcontact (void)
{
  long i;
  
  for (i=0;i<nb;i++){
    delete [] nip[i];
    delete [] intordsm[i];
  }
  delete [] nip;
  delete [] intordsm;

  delete [] ncomp;
  
}



/**
   The function initializes basic data on elements.
   
   @param eid - element id      
   
   JK, 11.6.2006

void plquadcontact::eleminit (long eid)
{

  long ii,jj;

  Mt->elements[eid].nb=nb;
  Mt->elements[eid].intordsm = new long* [nb];
  Mt->elements[eid].nip = new long* [nb];

  for (ii=0;ii<nb;ii++){
    Mt->elements[eid].intordsm[ii] = new long [nb];
    Mt->elements[eid].nip[ii] = new long [nb];
    for (jj=0;jj<nb;jj++){
      Mt->elements[eid].intordsm[ii][jj]=intordsm[ii][jj];
      Mt->elements[eid].nip[ii][jj]=nip[ii][jj];
    }
  }
}
*/


/**
   function approximates function defined by nodal values

   @param xi,eta - coordinates on element
   @param nodval - nodal values
   
   JK
*/
double plquadcontact::approx (double xi,vector &nodval)
{
  double f;
  vector bf(nne);
  
  bf_lin_1d (bf.a,xi);
  // we need approximation as on 1D element with 2 nodes
  bf[2] = 0.0;
  bf[3] = 0.0;
  
  scprd (bf,nodval,f);

  return f;
}



/**
   The function assembles transformation %matrix x_g = T x_l
   for transformation from the local nodal coordinate systems.
   
   @param nodes - array containing node numbers
   @param tmat - transformation %matrix
   
   JK, 11.6.2006
*/
void plquadcontact::transf_matrix (ivector &nod,matrix &tmat)
{
  long i,n,m;

  fillm (0.0,tmat);

  n=nod.n;
  m=tmat.m;
  for (i=0;i<m;i++){
    tmat[i][i]=1.0;
  }
  
  for (i=0;i<n;i++){
    if (Mt->nodes[nod[i]].transf>0){
      tmat[i*2][i*2]   = Mt->nodes[nod[i]].e1[0];    tmat[i*2][i*2+1]   = Mt->nodes[nod[i]].e2[0];
      tmat[i*2+1][i*2] = Mt->nodes[nod[i]].e1[1];    tmat[i*2+1][i*2+1] = Mt->nodes[nod[i]].e2[1];
    }
  }
}



/**
   The function computes stiffness %matrix of plane rectangular
   finite element for contact problems.
   
   @param eid - number of element
   @param ri,ci - row and column indices
   @param sm - resulting stiffness %matrix

   @return The function returns resulting stiffnes %matrix in the parameter sm.
   
   JF, 29.10.2012
*/
void plquadcontact::stiffness_matrix (long eid,long ri,long ci,matrix &sm)
{
  long i,ipp;
  double xi,jac,thick;
  ivector nodes (nne);
  vector w(intordsm[0][0]),gp(intordsm[0][0]);
  vector x(nne),y(nne), t(nne);
  matrix gm(ncomp[0],ndofe),d(tncomp,tncomp);

  fillm (0.0,sm);

  Mt->give_node_coord2d(x,y,eid);
  Mt->give_elemnodes (eid,nodes);
  Mc->give_thickness (eid,nodes,t);

  gauss_points (gp.a,w.a,intordsm[0][0]);

  ipp=Mt->elements[eid].ipp[ri][ci];
  
  for (i=0;i<intordsm[0][0];i++)
  {
    xi=gp[i];

    geom_matrix (gm, x, y, xi, jac);
    //  matrix of stiffness of the material
    Mm->matstiff (d,ipp);
    //  thickness in integration point
    thick = approx (xi,t);
    jac*=w[i]*thick;
    //  contribution to the stiffness matrix of the element
    bdbjac (sm,gm,d,gm,jac);

    ipp++;
  }
}



/**
  Function computes stiffness %matrix of one element. If it is required, nodal values 
  are transformed to the local nodal coordinate systems.

  @param eid - number of element
  @param sm - stiffness %matrix (output)

  @return The function returns required stiffness %matrix in the parameter sm.

  Created by Tomas Koudelka, 11.2008
*/
void plquadcontact::res_stiffness_matrix(long eid, matrix &sm)
{
  long transf;
  ivector nodes(nne);
  
  stiffness_matrix(eid,0,0,sm);
  
  //  transformation of stiffness matrix
  //  (in the case of nodal coordinate systems)
  Mt->give_elemnodes(eid,nodes);
  transf = Mt->locsystems(nodes);
  if (transf>0){
    matrix tmat(ndofe,ndofe);
    transf_matrix(nodes,tmat);
    glmatrixtransf(sm,tmat);
  }
}



/**
   The function computes strains at integration points. If there are 
   defined loacal nodal coordinate systems then the corresponding 
   components of the dislacement %vector are transformed.
   
   @param lcid - load case id
   @param eid - element id
   
   @return The function stores resulting strains in the array strains 
           of the element integration points.

   JK, 11.6.2006
*/
void plquadcontact::res_mainip_strains (long lcid,long eid)
{
 
  vector aux,x(nne),y(nne),r(ndofe);
  ivector nodes(nne);
  matrix tmat;

  //  nodal displacements
  eldispl (lcid,eid,r.a);
  
  //  transformation of displacement vector
  long transf = Mt->locsystems (nodes);
  if (transf>0){
    allocv (ndofe,aux);
    allocm (ndofe,ndofe,tmat);
    transf_matrix (nodes,tmat);
    //locglobtransf (aux,r,tmat);
    lgvectortransf (aux,r,tmat);
    copyv (aux,r);
    destrv (aux);
    destrm (tmat);
  }
  
  mainip_strains (lcid,eid,0,0,r);
  
}



/**
   The function computes strains at integration points of element.
   It is used in geometrically linear problems.
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   @param ii - number of block
   @param r - %vector of nodal displacements
   
   @return The function stores resulting strains in the array strains 
           of the element integration points.

   JK, 11.6.2006
*/
void plquadcontact::mainip_strains (long lcid,long eid,long ri,long ci,vector &r)
{
  long i,ipp;
  double xi,jac;
  ivector nodes (nne);
  vector w(intordsm[ri][ci]),gp(intordsm[ri][ci]),t(nne);
  vector x(nne),y(nne);
  matrix gm(ncomp[0],ndofe),d(tncomp,tncomp);
  vector eps(tncomp);

  fillm (0.0,gm);

  Mt->give_node_coord2d(x,y,eid);
  Mt->give_elemnodes (eid,nodes);

  gauss_points (gp.a,w.a,intordsm[ri][ci]);

  ipp=Mt->elements[eid].ipp[ri][ci];
  
  for (i=0;i<intordsm[ri][ci];i++)
  {
    xi=gp[i];

    geom_matrix (gm, x, y, xi, jac);

    mxv (gm, r, eps);

    Mm->storestrain (lcid,ipp,eps);

    ipp++;
  }  
}



/**
   The function assembles strain-displacement (geometric) %matrix.
   
   @param gm - resulting geometric %matrix (output)
   @param x,y - nodal coordinates
   @param xi -  natural coordinate of required integration point
   @param jac - Jacobian of transformation (output)
   
   @return The function returns resulting geometric %matrix in the parameter gm.

   JF, 15.10.2012
*/

void plquadcontact::geom_matrix (matrix &gm,vector &x,vector &y,double xi,double &jac)
{
  matrix transf(2,2);
  matrix n1(2,2),n2(2,2);
  matrix pom(2,2);
  double l,dx,dy;

  dx = x[1] - x[0];
  dy = y[1] - y[0];

  fillm(0.0,gm);

  fillm(0.0,n1);
  fillm(0.0,n2);
  fillm(0.0,transf);
  fillm(0.0,pom);

  n1[0][0]=(1-xi)/2;
  n1[1][1]=(1-xi)/2;
  
  n2[0][0]=(1+xi)/2;
  n2[1][1]=(1+xi)/2;

  l=sqrt(dx*dx+dy*dy);

  transf[0][0]=dx/l;
  transf[0][1]=dy/l;
  transf[1][0]=-dy/l;
  transf[1][1]=dx/l;

  mxm(transf,n1,pom);
  gm[0][4]=pom[0][0];
  gm[0][5]=pom[0][1];
  gm[1][4]=pom[1][0];
  gm[1][5]=pom[1][1];

  cmulm(-1.0,pom);
  gm[0][0]=pom[0][0];
  gm[0][1]=pom[0][1];
  gm[1][0]=pom[1][0];
  gm[1][1]=pom[1][1];

  mxm(transf,n2,pom);
  gm[0][6]=pom[0][0];
  gm[0][7]=pom[0][1];
  gm[1][6]=pom[1][0];
  gm[1][7]=pom[1][1];

  cmulm(-1.0,pom);
  gm[0][2]=pom[0][0];
  gm[0][3]=pom[0][1];
  gm[1][2]=pom[1][0];
  gm[1][3]=pom[1][1];

  jac=l/2;
}



/**
   The function computes stresses at element integration points
   
   @param lcid - load case id
   @param eid - element id
   
   @return The function stores resulting stresses in the stress array 
           of the element integration points.

   TKo, 11.2012
*/
void plquadcontact::res_mainip_stresses (long lcid,long eid)
{
  long ri,ci;
  ri=0;
  ci=0;
  compute_nlstress (lcid,eid,ri,ci);
}



/**
  Function computes correct stresses at integration points on element.

  @param lcid - number of load case
  @param eid - element id
  @param ri,ci - row and column indices

  @return The function stores resulting stresses in the stress array 
          of the element integration points.

  Created by Tomas Koudelka, 11.2008
*/

void plquadcontact::compute_nlstress (long lcid,long eid,long ri,long ci)
{
  long i,ipp;
  
  ipp=Mt->elements[eid].ipp[ri][ci];
  
  for(i=0;i<intordsm[0][0];i++)
  {
    //  computation of correct stresses
    if (Mp->strcomp==1)
      Mm->computenlstresses(ipp);
    ipp++;
  }
}



/**
  Function computes internal forces in the case of geometrical linear computation.
   
  @param lcid - number of load case
  @param eid - element id
  @param ri,ci - row and column indices
  @param ifor - %vector of internal forces (output)
  @param x,y - %vectors of nodal coordinates
   
  @return The function returns %vector of internal forces in the parameter ifor.

  Created by Tomas Koudelka, 11.2008
*/
void plquadcontact::internal_forces (long lcid,long eid,long ri,long ci,vector &ifor,vector &x, vector &y)
{
  integratedquant iq;
  iq=locstress;
  
  //  computation of stresses
  compute_nlstress (lcid,eid,ri,ci);
  
  //  integration of stresses over the element
  elem_integration (iq,lcid,eid,ri,ci,ifor,x,y);
}



/**
  Function computes resulting internal forces (from correct stresses)
   
  @param lcid - number of load case
  @param eid - element id
  @param ifor - %vector of internal forces
   
  @return The function returns resulting %vector of internal forces in the parameter ifor.

  Created by Tomas Koudelka, 11.2008
*/
void plquadcontact::res_internal_forces (long lcid,long eid,vector &ifor)
{
  long transf;
  ivector nodes (nne);
  vector v(ndofe),x(nne),y(nne);
  
  Mt->give_node_coord2d (x,y,eid);

  internal_forces (lcid,eid,0,0,ifor,x,y);
  
  //  transformation of nodal forces
  //  (in the case of nodal coordinate systems)
  Mt->give_elemnodes (eid,nodes);
  transf = Mt->locsystems (nodes);
  if (transf>0){
    matrix tmat (ndofe,ndofe);
    transf_matrix (nodes,tmat);
    //globloctransf (ifor,v,tmat);
    glvectortransf (ifor,v,tmat);
    copyv (v,ifor);
  }
}



/**
   The function integrates selected quantity on the selected finite element.
   It results in nodal values.
   
   @param iq - type of integrated quantity (see alias.h)
   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   @param nv - nodal values
   @param x,y - nodal coordinates

   @return The function returns nodal values of the integrated quantity in the parameter nv.
   
   JF, 29.10.2012
*/
void plquadcontact::elem_integration (integratedquant iq,long lcid,long eid,long ri,long ci,vector &nv,vector &x,vector &y)
{
  long ipp,i;
  double xi,jac,thick;
  ivector nodes (nne);
  vector ipv(tncomp),contr(ndofe), t(nne);
  vector w(intordsm[0][0]);
  matrix gm(tncomp,ndofe);
  vector gp(intordsm[0][0]);

  
  fillv(0.0,nv);
  
  gauss_points (gp.a,w.a,intordsm[0][0]);

  ipp=Mt->elements[eid].ipp[ri][ci];
  Mt->give_elemnodes (eid,nodes);
  Mc->give_thickness (eid,nodes,t);

  for (i=0;i<intordsm[0][0];i++)
  {
    xi  =gp[i];
    //  function assembles required quantity at integration point
    Mm->givequantity(iq,lcid,ipp,0,ipv);
    //  strain-displacement (geometric) matrix
    geom_matrix(gm,x,y,xi,jac);
    //  contribution to the internal forces
    mtxv(gm,ipv,contr);
    //  thickness in integration point
    thick = approx (xi,t);
    cmulv(jac*w[i]*thick,contr);
    //  summation
    addv(contr,nv,nv);
    ipp++;
  }
}





/*
   OLD VERSION
*/


// plquadcontact::plquadcontact (void)
// {
//   long i,j;

//   nne=4;  ndofe=8;
//   nb=1;  tncomp=2;

//   ncomp = new long [nb];
//   ncomp[0]=2;

//   nip = new long* [nb];
//   //intordsm = new long* [nb];
//   for (i=0;i<nb;i++){
//     nip[i] = new long [nb];
//     //intordsm[i] = new long [nb];
//   }
  
//   nip[0][0]=2;
  
//   tnip=0;
//   for (i=0;i<nb;i++){
//     for (j=0;j<nb;j++){
//       tnip+=nip[i][j];
//     }
//   }
// }

// plquadcontact::~plquadcontact (void)
// {
//   long i;
  
//   for (i=0;i<nb;i++){
//     delete [] nip[i];
//     //delete [] intordsm[i];
//   }
//   delete [] nip;
//   //delete [] intordsm;

//   delete [] ncomp;
  
// }

// /**
//    function initializes basic data on elements
   
//    @param eid - element id
   
//    JK, 11.6.2006
// */
// void plquadcontact::eleminit (long eid)
// {

//   long ii,jj;

//   Mt->elements[eid].nb=nb;
//   //Mt->elements[eid].intordsm = new long* [nb];
//   Mt->elements[eid].nip = new long* [nb];

//   for (ii=0;ii<nb;ii++){
//     //Mt->elements[eid].intordsm[ii] = new long [nb];
//     Mt->elements[eid].nip[ii] = new long [nb];
//     for (jj=0;jj<nb;jj++){
//       //Mt->elements[eid].intordsm[ii][jj]=intordsm[ii][jj];
//       Mt->elements[eid].nip[ii][jj]=nip[ii][jj];
//     }
//   }

// }

// /**
//    function assembles transformation %matrix x_g = T x_l
   
//    @param nodes - array containing node numbers
//    @param tmat - transformation %matrix
   
//    JK, 11.6.2006
// */
// void plquadcontact::transf_matrix (ivector &nod,matrix &tmat)
// {
//   long i,n,m;

//   fillm (0.0,tmat);

//   n=nod.n;
//   m=tmat.m;
//   for (i=0;i<m;i++){
//     tmat[i][i]=1.0;
//   }
  
//   for (i=0;i<n;i++){
//     if (Mt->nodes[nod[i]].transf>0){
//       tmat[i*2][i*2]   = Mt->nodes[nod[i]].e1[0];    tmat[i*2][i*2+1]   = Mt->nodes[nod[i]].e2[0];
//       tmat[i*2+1][i*2] = Mt->nodes[nod[i]].e1[1];    tmat[i*2+1][i*2+1] = Mt->nodes[nod[i]].e2[1];
//     }
//   }
// }
















// /**
//    function computes stiffness %matrix of plane rectangular
//    finite element for contact problems
   
//    @param eid - number of element
//    @param ri,ci - row and column indices
//    @param sm - stiffness %matrix
//    @param x,y - vectors of nodal coordinates
   
//    JK, 11.6.2006
// */
// void plquadcontact::stiffness_matrix (long eid,long ri,long ci,matrix &sm,vector &x,vector &y)
// {
//   long ipp;
//   matrix d(2,2);
  
//   fillm (0.0,sm);
  
//   //  number of appropriate integration point
//   ipp=Mt->elements[eid].ipp[ri][ci];

//   //  matrix of stiffness of the material
//   Mm->matstiff (d,ipp);
  
//   sm[0][0]=d[0][0];
//   sm[0][2]=0.0-d[0][0];

//   sm[1][1]=d[1][1];
//   sm[1][3]=0.0-d[1][1];

//   sm[2][0]=0.0-d[0][0];
//   sm[2][2]=d[0][0];
  
//   sm[3][1]=0.0-d[1][1];
//   sm[3][3]=d[1][1];
  


//   sm[4][4]=d[0][0];
//   sm[4][6]=0.0-d[0][0];

//   sm[5][5]=d[1][1];
//   sm[5][7]=0.0-d[1][1];

//   sm[6][4]=0.0-d[0][0];
//   sm[6][6]=d[0][0];

//   sm[7][5]=0.0-d[1][1];
//   sm[7][7]=d[1][1];

// }


// /**
//    function computes stiffness %matrix of quadrilateral element
//    for contact problems
   
//    @param lcid - load case id
//    @param eid - element id
//    @param sm - stiffness %matrix
   
//    JK, 11.6.2006
// */
// void plquadcontact::res_stiffness_matrix (long lcid,long eid,matrix &sm)
// {
//   long transf;
//   ivector nodes(nne);
//   vector x(nne),y(nne);
//   Mt->give_node_coord2d (x,y,eid);
//   Mt->give_elemnodes (eid,nodes);
  
//   stiffness_matrix (eid,0,0,sm,x,y);
  
//   //  transformation of stiffness matrix
//   transf = Mt->locsystems (nodes);
//   if (transf>0){
//     matrix tmat (ndofe,ndofe);
//     transf_matrix (nodes,tmat);
//     glmatrixtransf (sm,tmat);
//   }
// }



















// /**
//    function computes strains at integration points
   
//    @param lcid - load case id
//    @param eid - element id
   
//    JK, 11.6.2006
// */
// void plquadcontact::res_mainip_strains (long lcid,long eid)
// {
//   vector aux,x(nne),y(nne),r(ndofe);
//   ivector nodes(nne);
//   matrix tmat;
  
//   //  coordinates of nodes of element
//   Mt->give_node_coord2d (x,y,eid);
//   //  node numbers of element
//   Mt->give_elemnodes (eid,nodes);
//   //  nodal displacements
//   eldispl (lcid,eid,r.a);
  
//   //  transformation of displacement vector
//   long transf = Mt->locsystems (nodes);
//   if (transf>0){
//     allocv (ndofe,aux);
//     allocm (ndofe,ndofe,tmat);
//     transf_matrix (nodes,tmat);
//     //locglobtransf (aux,r,tmat);
//     lgvectortransf (aux,r,tmat);
//     copyv (aux,r);
//     destrv (aux);
//     destrm (tmat);
//   }
  
//   //
//   mainip_strains (lcid,eid,0,0,x,y,r);
  
// }

// /**
//    function computes strains at integration points of element
//    function is used in geometrically linear problems
   
//    @param lcid - load case id
//    @param eid - element id
//    @param ri,ci - row and column indices
//    @param ii - number of block
//    @param x,y - arrays with node coordinates
//    @param r - %vector of nodal displacements
   
//    JK, 11.6.2006
// */
// void plquadcontact::mainip_strains (long lcid,long eid,long ri,long ci,vector &x,vector &y,vector &r)
// {
//   long ipp;
//   vector displdiff(2);
  
//   ipp=Mt->elements[eid].ipp[ri][ci];
  
//   displdiff[0]=r[0]-r[2];
//   displdiff[1]=r[1]-r[3];
  
//   Mm->storestrain (lcid,ipp,displdiff);
//   ipp++;
  
//   displdiff[0]=r[6]-r[4];
//   displdiff[1]=r[7]-r[5];
  
//   Mm->storestrain (lcid,ipp,displdiff);
  
// }












// /**
//    function computes internal forces
   
//    @param lcid - number of load case
//    @param eid - element id
//    @param ri,ci - row and column indices
//    @param ifor - vector of internal forces
//    @param x,y - vectors of nodal coordinates
   
//    JK, 11.6.2006
// */
// void plquadcontact::internal_forces (long lcid,long eid,long ri,long ci,vector &ifor,vector &x,vector &y)
// {
//   long ipp;
//   vector nodfor (2);

//   //  number of apropriate integration point
//   ipp=Mt->elements[eid].ipp[ri][ci];
  
//   if (Mp->strcomp==1)
//     Mm->computenlstresses (ipp);
  
//   Mm->givestress (lcid,ipp,nodfor);
  
//   ifor[0]=nodfor[0];
//   ifor[1]=nodfor[1];
//   ifor[2]=0.0-nodfor[0];
//   ifor[3]=0.0-nodfor[1];

//   if (Mp->strcomp==1)
//     Mm->computenlstresses (ipp);
  
//   Mm->givestress (lcid,ipp,nodfor);
  
//   ifor[4]=0.0-nodfor[0];
//   ifor[5]=0.0-nodfor[1];
//   ifor[6]=nodfor[0];
//   ifor[7]=nodfor[1];

// }


// /**
//    function computes internal forces
   
//    @param lcid - load case id
//    @param eid - element id
//    @param ifor - %vector of internal forces
   
//    JK, 11.6.2006
// */
// void plquadcontact::res_internal_forces (long lcid,long eid,vector &ifor)
// {
//   vector x(nne),y(nne);

//   Mt->give_node_coord2d (x,y,eid);

//   internal_forces (lcid,eid,0,0,ifor,x,y);
// }

